// $Id: Dump.java 15896 2012-06-22 16:57:51Z mike $

import org.faceless.pdf2.*;
import java.io.*;
import java.awt.color.ICC_Profile;
import java.util.*;
import java.net.URL;
import java.awt.geom.*;
import java.awt.Shape;
import java.text.DecimalFormat;
import java.security.Security;
import java.security.Provider;
import java.security.KeyStore;
import java.security.cert.X509Certificate;

/**
 * Usage: java Dump [--out <filename>] [--profile full | basic]
 *                  [--password <password>] [--provider <providername>] file.pdf...
 *
 * This example takes a PDF file and dumps out as much information as possible
 * about it. It's very complete so may display a lot of information!
 *
 * The command line options are:
 * --out <filename>             Redirects output to a file. The file will be in UTF-8
 *                              format, unlike output to System.out which is escaped.
 * --profile full               Extract the Full OutputProfile rather than the basic one.
 * --keystore <keystore>        Specify the keystore (for use with public key encrypted docs)
 * --password <password>        Specify the open password for the document.
 * --provider <provider>        Use the specified Security Provider. Typical values for
 *                              this would be "BC" to use the BouncyCastle provider, "IAIK"
 *                              to use the IAIK provider, "SUN" for the default provider
 *                              supplied with Java, or the class name of a Provider.
 */
public class Dump
{
    public static void main(String[] args) throws Exception {
        if (args.length==0) {
            System.err.println("USAGE: java Dump [--out <filename>] [--keystore keystore] [--profile full | basic] [--password <password>] [--provider <providername>] file.pdf ...\n");
            System.exit(1);
        }

        PrintStream out = System.out;
        String filename = null, password = null, keystore = null;
        boolean fullprofile = false;
        for (int i=0;i<args.length;i++) {
            if (args[i].equals("--password")) {
                password = args[++i];
            } else if (args[i].equals("--keystore")) {
                keystore = args[++i];
            } else if (args[i].equals("--profile")) {
                fullprofile = args[++i].equals("full");
            } else if (args[i].equals("--provider")) {
                String provname = args[++i];
                if (provname.equals("BC")) {
                    provname = "org.bouncycastle.jce.provider.BouncyCastleProvider";
                } else if (provname.equals("IAIK")) {
                    provname = "iaik.security.provider.IAIK";
                }
                Provider provider = Security.getProvider(provname);
                if (provider==null) {
                    Class cl = null;
                    try {
                        cl = Class.forName(provname);
                    } catch (ClassNotFoundException e) {}
                    if (cl!=null && Provider.class.isAssignableFrom(cl)) {
                        provider = (Provider)cl.newInstance();
                    } else {
                        throw new IllegalArgumentException("No Security Provider \""+provname+"\"");
                    }
                }
                Security.insertProviderAt(provider, 1);
            } else if (args[i].equals("--out")) {
                if (out != System.out) {
                    out.close();
                    System.out.println("Created \""+filename+"\"");
                }
                filename = args[++i];
                out = new PrintStream(new FileOutputStream(filename), true, "UTF-8");
            } else {
                EncryptionHandler handler = null;
                if (keystore != null && password != null) {
                    KeyStore store = KeyStore.getInstance("JKS");
                    FileInputStream kin = new FileInputStream(new File(keystore));
                    store.load(kin, null);
                    kin.close();
                    handler = new PublicKeyEncryptionHandler(store, null, password.toCharArray());
                } else if (password != null) {
                    handler = new StandardEncryptionHandler();
                    ((StandardEncryptionHandler)handler).setUserPassword(password);
                }
                Dump dump = new Dump(args[i], handler, out);
                dump.unicode = filename!=null;
                dump.fullprofile = fullprofile;
                dump.go();
                out.flush();
            }
        }

        if (out != System.out) {
            out.close();
            System.out.println("Created \""+filename+"\"");
        }
    }

    //--------------------------------------------------------------

    private final PDF pdf;
    private final PrintStream out;
    private boolean unicode, fullprofile;

    Dump(String file, EncryptionHandler handler, PrintStream out) throws IOException {
        PDFReader rd;
        if (file.equals("-")) {
            rd = new PDFReader(System.in, handler);
        } else if (file.startsWith("http:") || file.startsWith("https:")) {
            rd = new PDFReader(new URL(file).openStream(), handler);
        } else {
            rd = new PDFReader(new File(file), handler);
        }
        this.pdf = new PDF(rd);
        this.out = out;
        out.println("DUMPING \""+file+"\" with PDF Library "+PDF.VERSION+"/"+System.getProperty("java.vendor")+" Java "+System.getProperty("java.version")+"/"+System.getProperty("os.name")+"\n----------------------------------------------------------------------\n");
    }

    void go() throws Exception {
        Printer printer = new Printer(out);
        getInfoDictionary(printer);
        getOptions(printer);
        OutputProfile profile = getOutputProfile(printer);
        getOutputIntent(printer, profile, "GTS_PDFX", "PDF/X");
        getOutputIntent(printer, profile, "GTS_PDFA1", "PDF/A");
        getEncryption(printer);
        getJavaScript(printer);
        getDocumentActions(printer);
        getMetaData(pdf.getMetaData(), printer);
        getBookmarks(printer);
        getNamedActions(printer);
        getEmbeddedFiles(printer);
        getPages(printer);
        getXFA(printer);
        getForm(printer);
    }

    void getOptions(Printer out) {
        final String[] keys = { "view.fullscreen", "view.displaydoctitle", "view.hidetoolbar", "view.hidemenubar", "view.hidewindowui", "view.fitwindow", "view.centerwindow", "pagelayout", "pagemode", "view.area", "view.clip", "print.area", "print.clip", "print.scaling", "print.duplex", "print.matchtraysize", "print.numcopies", "print.pagerange" };

        out.startSection("Options");
        for (int i=0;i<keys.length;i++) {
            String key = keys[i];
            Object value = pdf.getOption(key);
            if (value!=null) {
                if (key=="print.pagerange") {
                    List pagelist = (List)value;
                    for (int j=0;j<pagelist.size();j++) {
                        PDFPage page = (PDFPage)pagelist.get(j);
                        pagelist.set(j, new Integer(page.getPageNumber()));
                    }
                }
                out.println(keys[i]+":\t"+value);
            }
        }
        out.endSection();
    }

    void getEncryption(Printer out) {
        out.startSection("Encryption");
        EncryptionHandler handler = pdf.getEncryptionHandler();
        if (handler!=null) {
            if (handler instanceof StandardEncryptionHandler || handler instanceof PublicKeyEncryptionHandler) {
                int change, print, extract;
                if (handler instanceof StandardEncryptionHandler) {
                    int version = ((StandardEncryptionHandler)handler).getVersion();
                    out.println("Standard PDF Encryption version "+version+": "+((StandardEncryptionHandler)handler).getDescription());
                    change = ((StandardEncryptionHandler)handler).getChange();
                    print = ((StandardEncryptionHandler)handler).getPrint();
                    extract = ((StandardEncryptionHandler)handler).getExtract();
                } else {
                    out.println("Public Key Encryption");
                    change = ((PublicKeyEncryptionHandler)handler).getChange();
                    print = ((PublicKeyEncryptionHandler)handler).getPrint();
                    extract = ((PublicKeyEncryptionHandler)handler).getExtract();
                }
                boolean change_forms = (change & StandardEncryptionHandler.CHANGE_FORMS)==StandardEncryptionHandler.CHANGE_FORMS;
                boolean change_layout = (change & StandardEncryptionHandler.CHANGE_LAYOUT)==StandardEncryptionHandler.CHANGE_LAYOUT;
                boolean change_annots = (change & StandardEncryptionHandler.CHANGE_ANNOTATIONS)==StandardEncryptionHandler.CHANGE_ANNOTATIONS;
                boolean change_none = !(change_forms | change_layout | change_annots);
                out.println("CanChange:\t"+(change_none?"None ":"")+(change_layout?"Assembly ":"")+(change_forms?"Forms ":"")+(change_annots?"Comments ":"")+" "+change);
                out.println("CanPrint:\t"+(print==StandardEncryptionHandler.PRINT_NONE ? "None" : print==StandardEncryptionHandler.PRINT_LOWRES ? "Low-res" : "High-res")+" "+print);
                out.println("CanExtract:\t"+(extract==StandardEncryptionHandler.EXTRACT_NONE ? "None" : extract==StandardEncryptionHandler.EXTRACT_ACCESSIBILITY ? "Accessibility" : "All")+" "+extract);
            } else {
                out.println("Non-standard Encryption Handler: "+handler.getFilterName());
            }
        }
        out.endSection();
    }

    OutputProfile getOutputProfile(Printer out) {
        out.startSection("Output Profile");
        OutputProfile profile = fullprofile ? pdf.getFullOutputProfile() : pdf.getBasicOutputProfile();
        for (int i=0;i<OutputProfile.Feature.ALL.length;i++) {
            OutputProfile.Feature feature = OutputProfile.Feature.ALL[i];
            if (profile.isSet(feature)) out.println("\""+feature+"\" is set");
        }
        Set fonts = profile.getFontInfo();
        if (fonts!=null) {
            out.startSection("Fonts");
            for (Iterator i = fonts.iterator();i.hasNext();) {
                OutputProfile.FontInfo fontinfo = (OutputProfile.FontInfo)i.next();
                out.println(fontinfo.getBaseName()+": "+(fontinfo.isCIDFont() ? "CID ":" Simple ")+fontinfo.getType());
            }
            out.endSection();
        }
        out.endSection();
        return profile;
    }

    void getOutputIntent(Printer out, OutputProfile profile, String type, String name) {
        if (profile.getOutputConditionIdentifier(type)!=null) {
            out.startSection(name+" Output Intent");
            out.println("Identifier:\t"+profile.getOutputConditionIdentifier(type));
            out.println("Condition:\t"+profile.getOutputCondition(type));
            out.println("Registry:\t"+profile.getOutputConditionRegistry(type));
            out.println("Info:    \t"+profile.getOutputIntentInfo(type));
            ICC_Profile icc = profile.getOutputIntentDestinationProfile(type);
            if (icc==null) {
                out.println("Profile: \tnot embedded");
            } else {
                out.println("Profile: \tembedded ("+icc.getNumComponents()+" components)");
            }
            out.endSection();
        }
    }

    void getJavaScript(Printer out) {
        out.startSection("JavaScript");
        String js = pdf.getJavaScript();
        if (js!=null) {
            out.println(js);
        }
        out.endSection();
    }

    void getBookmarks(Printer out) {
        out.startSection("Bookmarks");
        getBookmarks(out, pdf.getBookmarks(), "... ");
        out.endSection();
    }

    private void getBookmarks(Printer out, List l, String indent) {
        for (int i=0;l!=null && i<l.size();i++) {
            PDFBookmark b = (PDFBookmark)l.get(i);
            out.println(indent+b.getName()+":\t"+getActionBrief(b.getAction()));
            getBookmarks(out, b.getBookmarks(), "..."+indent);
        }
    }

    private String getActionBrief(PDFAction action) {
        if (action == null) {
            return "null";
        } else {
            String s = actionToString(action).replaceAll("\n", "\\\\n");
            if (s.length() > 60) {
                s = s.substring(0, 57) + "...";
            }
            return s;
        }
    }

    void getAction(PDFAction action, String name, Printer out) {
        if (action != null) {
            out.startSection(name);
            while (action!=null) {
                out.println(actionToString(action));
                action = action.getNext();
            }
            out.endSection();
        }
    }

    private String actionToString(PDFAction action) {
        if (action == null) {
            return "null";
        }
        String type = action.getType();
        if (type.equals("FormJavaScript")) {
            return action.getJavaScript();
        } else if (type.equals("GoToR")) {
            return "Remote: "+action.getRemoteFilename()+" page "+action.getRemotePageNumber();
        } else if (type.equals("FormReset") || type.equals("FormSubmit")) {
            String desc = type+":";
            if (type.equals("FormSubmit")) {
                desc += " URL="+action.getURL();
                int method = action.getFormSubmitMethod();
                if (method==PDFAction.METHOD_HTTP_POST) {
                    desc += " method=POST";
                } else if (method==PDFAction.METHOD_HTTP_IMAGEMAP_POST) {
                    desc += " method=POST(Imagemap)";
                } else if (method==PDFAction.METHOD_FDF) {
                    desc += " method=FDF";
                } else if (method==PDFAction.METHOD_XML) {
                    desc += " method=XML";
                } else if (method==PDFAction.METHOD_FDF_WITH_ALL_ANNOTATIONS) {
                    desc += " method=FDF(include all annotations)";
                } else if (method==PDFAction.METHOD_XML_WITH_ALL_ANNOTATIONS) {
                    desc += " method=XML(include all annotations)";
                } else if (method==PDFAction.METHOD_FDF_WITH_MY_ANNOTATIONS) {
                    desc += " method=FDF(include my annotations)";
                } else if (method==PDFAction.METHOD_PDF) {
                    desc += " method=PDF";
                } else {
                    desc += " method=Unknown!";
                }
                int flags = action.getFormSubmitFlags();
                if ((flags & PDFAction.METHOD_FLAG_EMPTYFIELDS)!=0) {
                    desc +=" emptyfields=true";
                }
                if ((flags & PDFAction.METHOD_FLAG_CANONICALDATES)!=0) {
                    desc +=" canonicaldates=true";
                }
                desc += " flags=0x"+Integer.toHexString(flags);
            }
            Collection c = action.getFormSubmitFields();
            String fieldnames;
            if (c.size() == pdf.getForm().getElements().size()) {
                desc += " fields=<all>";
            } else {
                List l = new ArrayList();
                for (Iterator i = c.iterator();i.hasNext();) {
                    FormElement e = (FormElement)i.next();
                    l.add(e.getForm().getName(e));
                }
                desc += " fields="+l;
            }
            return desc;
        } else if (type.equals("GoToURL")) {
            return "URL: "+action.getURL();
        } else if (type.startsWith("GoTo")) {
            float[] coords = action.getGoToCoordinates();
            String stringcoords=null;
            if (coords==null) {
                stringcoords=null;
            } else if (coords.length==1) {
                stringcoords = "["+coords[0]+"]";
            } else if (coords.length==2) {
                stringcoords = "["+coords[0]+" "+coords[1]+"]";
            } else if (coords.length==3) {
                stringcoords = "["+coords[0]+" "+coords[1]+" "+coords[2]+"]";
            } else if (coords.length==4) {
                stringcoords = "["+coords[0]+" "+coords[1]+" "+coords[2]+" "+coords[3]+"]";
            }
            PDFPage page = action.getPage();
            return type+": page="+(page==null?"nopage":Integer.toString(page.getPageNumber()))+" coords="+stringcoords;
        } else {
            return action.toString();
        }
    }

    void getAnnotation(PDFAnnotation annot, String title, Printer out) throws IOException {
        out.startSection(title);
        out.println("Type:\t"+annot.getType());
        float[] r = annot.getRectangle();
        out.println("Rect:\t"+fmt(r[0])+", "+fmt(r[1])+" - "+fmt(r[2])+", "+fmt(r[3]));
        if (annot.getUniqueID()!=null) out.println("Unique ID:\t"+annot.getUniqueID());
        if (annot.getModifyDate()!=null) out.println("Modified:\t"+annot.getModifyDate().getTime());
        if (annot.getCreationDate()!=null) out.println("Created:\t"+annot.getCreationDate().getTime());
        if (annot.isReadOnly()) out.println("Read-Only:\ttrue");
        if (annot.isPositionLocked() || annot.isContentLocked()) out.println("Locked:\t"+(annot.isPositionLocked() ? "position ":"")+(annot.isContentLocked()?"content":""));
        if (!annot.isVisible()) out.println("Invisible:\ttrue");
        if (!annot.isPrintable()) out.println("Printable:\tfalse");
        if (annot.getContents()!=null) {
            out.println("");
            out.println(fmt(annot.getContents().trim()));
        }
        if (annot instanceof AnnotationFile) {
            getEmbeddedFile("File", ((AnnotationFile)annot).getFile(), out);
        }
        if (annot instanceof WidgetAnnotation) {
            out.println("Field:\t"+pdf.getForm().getName(((WidgetAnnotation)annot).getField()));
            for (int i=0;i<Event.ALL.length;i++) {
                PDFAction act = ((WidgetAnnotation)annot).getAction(Event.ALL[i]);
                getAction(act, Event.ALL[i].toString(), out);
            }
        } else {
            if (annot instanceof AnnotationLink) {
                getAction(((AnnotationLink)annot).getAction(), "Action", out);
                float[] corners = ((AnnotationLink)annot).getCorners();
                if (corners!=null) {
                    StringBuffer sb = new StringBuffer();
                    for (int i=0;i<corners.length;i+=2) {
                        sb.append(fmt.format(corners[i])+","+fmt.format(corners[i+1])+" ");
                        if (i%8==6) sb.append("  ");
                    }
                    out.println("Corners:\t"+sb);
                }
            } else if (annot instanceof AnnotationMarkup) {
                float[] corners = ((AnnotationMarkup)annot).getCorners();
                if (corners!=null) {
                    StringBuffer sb = new StringBuffer();
                    for (int i=0;i<corners.length;i+=2) {
                        sb.append(fmt.format(corners[i])+","+fmt.format(corners[i+1])+" ");
                        if (i%8==6) sb.append("  ");
                    }
                    out.println("Corners:\t"+sb);
                }
            } else if (annot instanceof AnnotationShape) {
                Shape shape = ((AnnotationShape)annot).getShape();
                out.println("Shape:\t"+shape);
                out.println("Style:\t"+((AnnotationShape)annot).getStyle());
            }
            if (annot.getSubject()!=null) out.println("Subject:\t"+annot.getSubject());
            if (annot.getAuthor()!=null) out.println("Author:\t"+annot.getAuthor());
        }
        getMetaData(annot.getMetaData(), out);
        List l = annot.getReviews();
        for (int i=0;i<l.size();i++) {
            AnnotationNote note = (AnnotationNote)l.get(i);
            out.println(" * "+note.getStatus()+" set by "+note.getAuthor()+" at "+note.getModifyDate().getTime());
        }
        l = annot.getReplies();
        for (int i=0;i<l.size();i++) {
            getAnnotation((PDFAnnotation)l.get(i), "Reply "+(i+1)+"/"+l.size(), out);
        }
        out.endSection();
    }

    void getInfoDictionary(Printer out) {
        out.startSection("Info Dictionary");
        out.println("DocumentID:\t"+pdf.getDocumentID(true)+" "+pdf.getDocumentID(false));
        out.println("Revisions:\t"+pdf.getNumberOfRevisions());
        for (Iterator i = pdf.getInfo().entrySet().iterator();i.hasNext();) {
            Map.Entry e = (Map.Entry)i.next();
            if (!(e.getValue() instanceof Calendar)) {
                out.println(e.getKey()+":\t"+fmt(e.getValue().toString()));
            }
        }
        if (pdf.getLocale()!=null) out.println("Locale:\t"+pdf.getLocale());
        out.endSection();
    }

    void getDocumentActions(Printer out) {
        out.startSection("Document Actions");
        for (int i=0;i<Event.ALL.length;i++) {
            PDFAction act = pdf.getAction(Event.ALL[i]);
            getAction(act, Event.ALL[i].toString(), out);
        }
        out.endSection();
    }

    void getMetaData(Reader metadata, Printer out) {
        out.startSection("MetaData");
        if (metadata!=null) {
            StringBuffer sb = new StringBuffer();
            try {
                BufferedReader br = new BufferedReader(metadata);
                String s;
                while ((s=br.readLine())!=null) sb.append(s+"\n");
            } catch (IOException e) { }
            out.println(sb.toString().trim());
        }
        out.endSection();
    }

    void getNamedActions(Printer out) {
        out.startSection("Named Actions");
        Map actions = pdf.getNamedActions();
        for (Iterator i = actions.keySet().iterator();i.hasNext();) {
            String name = (String)i.next();
            PDFAction action = (PDFAction)actions.get(name);
            getAction(action, name, out);
        }
        out.endSection();
    }

    void getEmbeddedFiles(Printer out) throws IOException {
        out.startSection("Embedded Files");
        Map files = pdf.getEmbeddedFiles();
        for (Iterator i = files.keySet().iterator();i.hasNext();) {
            String name = (String)i.next();
            getEmbeddedFile(name, (EmbeddedFile)files.get(name), out);
        }
        out.endSection();
    }

    void getEmbeddedFile(String handle, EmbeddedFile file, Printer out) throws IOException {
        out.startSection(handle);
        out.println("Name\t"+file.getName());
        if (file.getType()!=null) out.println("Type\t"+file.getType());
        if (file.getModDate()!=null) out.println("Modified\t"+file.getModDate());
        if (file.getCreationDate()!=null) out.println("Created\t"+file.getCreationDate());
        if (file.getSize()!=-1) out.println("Size\t"+file.getSize());
        if (file.getDescription()!=null) out.println("Description\t"+file.getDescription());
        out.println("Checksum\t"+(file.isValid() ? "Probably OK" : "Definitely corrupt"));
        out.endSection();
    }

    void getPages(Printer out) throws Exception {
        out.startSection("Pages");
        List l = pdf.getPages();
        for (int i=0;i<l.size();i++) {
            PDFPage page = (PDFPage)l.get(i);
            getPage(page, out);
        }
        out.endSection();
    }

    void getPage(PDFPage page, Printer out) throws Exception {
        String label = pdf.getPageLabel(page.getPageNumber() - 1);
        out.startSection("Page "+page.getPageNumber()+"/"+pdf.getNumberOfPages() + (label == null ? "" : "  ("+label+")"));
        out.println("Page Size:\t"+page.getWidth()+" x "+page.getHeight());
        String[] boxes = { "CropBox", "ArtBox", "BleedBox", "TrimBox" };
        for (int i=0;i<boxes.length;i++) {
            float[] box = page.getBox(boxes[i]);
            if (box!=null) {
                out.println(boxes[i]+":\t"+fmt(box[0])+", "+fmt(box[1])+" - "+fmt(box[2])+", "+fmt(box[3]));
            }
        }
        int rotate = page.getPageOrientation();
        if (rotate!=0) {
            out.println("Orientation:\t"+rotate);
        }
        for (int i=0;i<Event.ALL.length;i++) {
            PDFAction act = page.getAction(Event.ALL[i]);
            getAction(act, Event.ALL[i].toString(), out);
        }

        List annots = page.getAnnotations();
        for (int j=0;j<annots.size();j++) {
            PDFAnnotation annot = (PDFAnnotation)annots.get(j);
            if (annot.getInReplyTo()==null && !annot.getType().equals("Popup")) {
                getAnnotation(annot, "Annotation "+(j+1)+"/"+annots.size(), out);
            }
        }
        getMetaData(page.getMetaData(), out);
        out.endSection();
    }

    /**
     * Dump information on the XFA form. This method requires Java 1.4,
     * specifically the javax.xml.transform package. If you're still
     * running Java 1.3 you can comment it out and the call to it.
     */
    void getXFA(Printer out) throws Exception {
        out.startSection("XFA Form");
        InputStream in = pdf.getForm().getXFAElement(null);
        if (in!=null) {
            javax.xml.transform.Transformer serializer = javax.xml.transform.TransformerFactory.newInstance().newTransformer();
            serializer.setOutputProperty(javax.xml.transform.OutputKeys.INDENT, "yes");
            serializer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
            StringWriter s = new StringWriter();
            serializer.transform(new javax.xml.transform.stream.StreamSource(in), new javax.xml.transform.stream.StreamResult(s));
            out.println(s.toString());
        }
        out.endSection();
    }

    void getForm(Printer out) throws Exception {
        out.startSection("Form");
        Map form = pdf.getForm().getElements();

        for (Iterator i=form.keySet().iterator(); i.hasNext();) {
            String name=(String)i.next();
            FormElement el = pdf.getForm().getElement(name);
            if (el instanceof FormSignature) {
                getFormSignature((FormSignature)el, name, out);
            } else if (el instanceof FormText) {
                getFormText((FormText)el, name, out);
            } else if (el instanceof FormButton) {
                getFormButton((FormButton)el, name, out);
            } else if (el instanceof FormCheckbox || el instanceof FormRadioButton) {
                getFormRadioBox(el, name, out);
            } else if (el instanceof FormChoice) {
                getFormChoice((FormChoice)el, name, out);
            } else if (el instanceof FormBarCode) {
                getFormBarCode((FormBarCode)el, name, out);
            } else {
                out.startSection("Unknown field type for field "+name);
            }
            if (el.getDescription()!=null) out.println("Description:\t"+el.getDescription());
            if (el.isReadOnly()) out.println("Read-Only:\ttrue");
            if (el.isRequired()) out.println("Required:\ttrue");
            if (!el.isSubmitted()) out.println("Submitted:\tfalse");
            for (int j=0;j<Event.ALL.length;j++) {
                PDFAction act = el.getAction(Event.ALL[j]);
                getAction(act, Event.ALL[j].toString(), out);
            }
            out.endSection();
        }
        out.endSection();
    }

    void getFormSignature(FormSignature sig, String name, Printer out) throws Exception {
        String[] certifications = { "No", "Yes: No Changes", "Yes: Modify Forms", "Yes: Modify Comments" };
        out.startSection("Signature \""+name+"\"");
        if (sig.getState()==sig.STATE_BLANK) {
            out.println("State:\tBlank");
        } else {
            KeyStore keystore = FormSignature.loadDefaultKeyStore();
            Calendar when = sig.getSignDate();
            out.println("Type:\t"+sig.getFilter());
            Object[] o = getSignatureValidity(sig, keystore);
            String validity = (String)o[0];
            X509Certificate cert = (X509Certificate)o[1];
            if (when != null) { // Only null for non-standard SignatureHandlers
                String timestampvalidity = getSignatureTimeStampValidity(sig, keystore);
                out.println("Signed:\t"+when.getTime()+" ("+timestampvalidity+")");
            }
            out.println("Certified:\t"+certifications[sig.getCertificationType()]);
            if (sig.getName()!=null) out.println("Signed by:\t"+sig.getName());
            if (sig.getReason()!=null) out.println("Reason:\t"+sig.getReason());
            if (sig.getLocation()!=null) out.println("Location:\t"+sig.getLocation());
            out.println("Validity:\t"+validity);
            if (cert!=null) {
                out.startSection("Unverifiable Certificate");
                out.println(cert.toString());
                out.endSection();
            }
        }
    }

    /**
     * Verify the signature.
     *
     * Unfortunately we want to return two things here - the "validity" string
     * and the unverifiable X.509 certificate. We return them in an array...
     */
    private Object[] getSignatureValidity(FormSignature sig, KeyStore keystore) {
        boolean verified=false, verifiable=false, certsok=false;
        X509Certificate cert = null;
        Calendar when = sig.getSignDate();
        try {
            verified = sig.verify();
            verifiable = true;
            if (sig.getSignatureHandler() instanceof PKCS7SignatureHandler) {
                PKCS7SignatureHandler handler = (PKCS7SignatureHandler)sig.getSignatureHandler();
                X509Certificate[] certs = handler.getCertificates();
                System.out.println("Found "+certs.length+" certs");
                cert = FormSignature.verifyCertificates(certs, keystore, null, when);
                certsok = cert==null;
            } else if (sig.getSignatureHandler() instanceof DSE200Handler) {
            } else {
            }
        } catch (Exception e) {
            System.out.println("ERROR: Verifying certificate");
            e.printStackTrace();
        }
        String validity="";
        if (sig.getNumberOfRevisionsCovered()!=pdf.getNumberOfRevisions()) {
            if (sig.getNumberOfRevisionsCovered()==0) {
                validity = "Bad - signature doesn't cover an entire revision. The section that is covered is ";
            } else {
                validity = "Partial - covers the first "+sig.getNumberOfRevisionsCovered()+" of "+pdf.getNumberOfRevisions()+" revisions. Those covered are ";
            }
        }
        if (verified && verifiable && certsok) {
            validity += "Perfect (signature and certificates verified)";
        } else if (verified && verifiable && !certsok) {
            validity += "Good (signature verified, but not certificates)";
        } else if (verifiable) {
            validity += "Bad (signature not verified - document has been altered)";
        } else {
            validity += "Unknown (not a PKCS#7 Signature, unable to verify)";
        }
        return new Object[] { validity, cert };
    }

    /**
     * Verify the RFC3161 TimeStamp on the signature
     * if there is one (only on PKCS#7 sigs).
     */
    private String getSignatureTimeStampValidity(FormSignature sig, KeyStore keystore) {
        X509Certificate[] certs = null;
        if (sig.getSignatureHandler() instanceof PKCS7SignatureHandler) {
            try {
                certs = ((PKCS7SignatureHandler)sig.getSignatureHandler()).getTimeStampCertificates();
            } catch (Exception e) {
                System.out.println("ERROR: Verifying TimeStamp certificates");
                e.printStackTrace();
            }
        }

        String validity;
        if (certs!=null) {
            Calendar when = sig.getSignDate();
            String issuer = null;
            X509Certificate badcert = null;
            try {
                issuer = FormSignature.getSubjectField(certs[0], "CN");
                badcert = FormSignature.verifyCertificates(certs, keystore, null, when);
            } catch (Exception e) {
                System.out.println("ERROR: Verifying certificate");
                e.printStackTrace();
            }
            if (badcert==null) {
                validity = "guaranteed by verified timestamp from \""+issuer+"\")";
            } else {
                validity = "guaranteed by unknown timestamp from \""+issuer+"\")";
            }
        } else {
            validity = "from computer clock";
        }
        return validity;
    }

    void getFormText(FormText text, String name, Printer out) throws Exception {
        out.startSection("Text \""+name+"\"");
        out.println("Value:\t"+(text.getValue()==null ? null : fmt(text.getValue())));
        if (text.getDefaultValue()!=null) out.println("DefaultValue:\t"+fmt(text.getDefaultValue()));
        if (text.getType()==FormText.TYPE_PASSWORD) out.println("Type:\t\tPassword");
        if (text.getType()==FormText.TYPE_BARCODE) out.println("Type:\t\tXFA Barcode");
        if (text.getType()==FormText.TYPE_MULTILINE) out.println("Type:\t\tmulti-line");
        if (text.getMaxLength()!=0) out.println("Max-Length:\t"+text.getMaxLength());
        if (text.getNumberOfCombs()!=0) out.println("Combs:\t"+text.getNumberOfCombs());
    }

    void getFormButton(FormButton button, String name, Printer out) throws Exception {
        out.startSection("Button \""+name+"\"");
        out.println("Buttons have no attributes");
    }

    void getFormRadioBox(FormElement box, String name, Printer out) throws Exception {
        out.startSection((box instanceof FormRadioButton ? "RadioButton" : "Checkbox")+" \""+name+"\"");
        if (box.getValue()!=null) out.println("Value:\t"+box.getValue());
        List annots = box.getAnnotations();
        for (int i=0;i<annots.size();i++) {
            WidgetAnnotation w = (WidgetAnnotation)annots.get(i);
            out.println(w.getValue());
        }
    }

    void getFormChoice(FormChoice choice, String name, Printer out) throws Exception {
        out.startSection("Choice \""+name+"\"");
        int type = choice.getType();
        out.println("Type:\t"+(type==FormChoice.TYPE_COMBO ? "Combo" : type==FormChoice.TYPE_SCROLLABLE ? "Scrollable" : type==FormChoice.TYPE_MULTISCROLLABLE ? "Multi-choice Scrollable" : "DropDown"));
        if (choice.getValue()!=null) out.println("Value:\t"+choice.getValue());

        Map opts = choice.getOptions();
        for (Iterator j=opts.entrySet().iterator();j.hasNext();) {
            Map.Entry e = (Map.Entry)j.next();
            String key = (String)e.getKey();
            String val = (String)e.getValue();
            out.println(" * "+key+" = "+val);
        }
    }
    
    void getFormBarCode(FormBarCode barcode, String name, Printer out) throws Exception {
        out.startSection("Barcode \""+name+"\"");
        out.println("Symbology:\t"+barcode.getSymbology()+" (with ECC="+barcode.getECC()+" and X-unit of "+fmt.format(barcode.getSymbolSize())+"mm"+(barcode.getCompression()!=0?", flate compressed":"")+")");
        if (barcode.getValue()!=null) out.println("Value:\t"+fmt(barcode.getValue()));
    }

    //-------------------------------------------------------------------------------------
    //
    // Remainder are utility methods and a subclass (Printer) for formatting the output
    //
    //-------------------------------------------------------------------------------------

    private static final DecimalFormat fmt = new DecimalFormat("#0.00");

    private static String fmt(double f) {
        if (f != f) {
            return "NaN";
        } else if (f == Double.POSITIVE_INFINITY) {
            return "+Inf";
        } else if (f == Double.POSITIVE_INFINITY) {
            return "-Inf";
        }  else {
            return f==(int)f ? Integer.toString((int)f) : fmt.format(f);
        }
    }

    private String fmt(String in) {
        StringBuffer st = new StringBuffer(in.length());
        for (int i=0;i<in.length();i++) {
            char c = in.charAt(i);
            if (c=='\n') {
                st.append("\\n");
            } else if (c=='\r') {
                st.append("\\r");
            } else if (c=='\t') {
                st.append("\\t");
            } else if ((c>=32 && c<=255) || unicode) {
                st.append(c);
            } else {
                String z = "0000"+Integer.toHexString(c);
                st.append("\\u"+z.substring(z.length()-4, z.length()));
            }
        }
        return st.toString();
    }

    /** 
     * Utility class to format output nicely
     */
    static class Printer {
        private PrintStream writer;
        private int printdepth, endprintdepth;
        private List headers;
        private boolean nl = true;
        private String prefix = "";
        static final String LINE = "---------------------------------------------------------------";

        Printer(PrintStream writer) {
            this.writer = writer;
            this.headers = new ArrayList();
        }

        void startSection(String header) {
            headers.add(header);
        }

        void endSection() {
            if (printdepth==headers.size()) {
                if (printdepth==1) {
                    writer.println();
                } else {
                    String header = (String)headers.get(headers.size()-1);
                    prefix = prefix.substring(0, prefix.length()-2);
                    writer.println(prefix+"+"+LINE);
                    endprintdepth = printdepth;
                }
                printdepth--;
            }
            headers.remove(headers.size()-1);
        }

        void println(StringBuffer s) {
            println(s.toString());
        }

        void print(StringBuffer s) {
            print(s.toString());
        }

        void println(String s) {
            print(s+"\n");
        }

        void print(String s) {
            while (printdepth < headers.size()) {
                String header = (String)headers.get(printdepth);
                if (printdepth==0) {
                    writer.println("=> "+header);
                    writer.println("-"+LINE);
                } else {
                    if (printdepth+1!=endprintdepth) writer.println(prefix+"+"+LINE);
                    prefix += "| ";
                    writer.println(prefix+"=> "+header);
                }
                printdepth++;
            }
            if (nl) writer.print(prefix);
            writer.print(s);
            nl = s.endsWith("\n");
        }
    }
}
