// $Id: PDFViewport.java 10851 2009-09-08 15:10:31Z mike $

package org.faceless.pdf2.viewer;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.print.*;
import java.io.*;
import java.util.*;
import javax.print.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.event.MouseEvent;
import org.faceless.pdf2.*;
import org.faceless.pdf2.Event;

/**
 * <p>
 * This class represents the "viewport" or "porthole" through which you see
 * the underlying {@link PDF} object. When you scroll a {@link PDFScrollPane}, what
 * moves is the viewport, like peering through a camera's viewfinder. Moving the
 * viewfinder upwards brings new things into view at the top of the picture and
 * loses things that were at the bottom.
 * <p></p>
 * Since 2.7.1 this class may be extended - the {@link #annotateComponent} method
 * can be overridden to add graphics operations to the page.
 * </p>
 *
 * @see PDFScrollPane
 * @since 2.5
 */
public class PDFViewport extends JPanel implements MouseMotionListener, MouseListener, KeyListener
{
    private PDF pdf;                            // The PDF...
    private PDFParser parser;                   // it's Parser...
    private PagePainter[] painters;             // and it's PagePainters.
    private BufferedImage pageImage;            // The section of page we're drawing
    private BufferedImage fullImage;            // The full page. May not be used

    private boolean updateImage;                // Whether we need to redraw
    private float dpi;                          // The DPI
    private float x1, y1, x2, y2;               // Scope of the viewport (points). 0,0 is top left
    private int pagenum;                        // The current pagenumber, starts at 0
    private float pagewidth, pageheight;        // Width/Height of the page (points)
    private float pageleft, pagebottom;         // Offset to the left/bottom of the page (points)

    private final Collection observers;         // List of viewport observers
    private PageWorker task;                    // The SwingWorker that redraws the page.
    private RenderingHints hints;
    private EventHandler eventhandler;          // The handler for mouse/keyboard events
    private final Map annotationmap;            // Map containing a map of shapes -> annotations

    // If the number of pixels required to render the image is less
    // than this value, the entire PDF is rendered. Above this size,
    // only sections of the page are rendered and the Viewport may
    // have to be redrawn if the viewport is scrolled.
    //
    private static final int MAXPIXELS = 2250000; // 1500x1500

    // The native DPI of the viewport. 72 may not always be right but that's
    // what we've chosen.
    //
    private static final float SCREENDPI = 72f;

    // The ColorModel used to draw the PDF.
    // * ColorModel.getRGBDefault() works on everything but Java 1.4 on OS X.
    // * PDFParser.RGB only works on Java 1.4 on OS X
    //
    private static final ColorModel colormodel = ColorModel.getRGBdefault();

    protected PDFViewport() {
        super();
        setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));
        setBackground(Color.GRAY);
        setDoubleBuffered(true);
        observers = new ArrayList();
        pagenum = -1;
        dpi = SCREENDPI;
        annotationmap = new HashMap();
        setEventHandler(new DefaultEventHandler());
    }

    /**
     * Return true if the zoom level is above the threshold we've chosen to
     * display the full image. If this method returns false, only a portion
     * of the page should be rendered.
     */
    private boolean isFullImageOK() {
        return pagewidth*pageheight*dpi/SCREENDPI <= MAXPIXELS;
    }

    /**
     * Shortcut for min <= val <= max
     */
    private static final float minmax(double min, double val, double max) {
        if (min>max) throw new IllegalArgumentException("min>max");
        return (float)(val < min ? min : val > max ? max : val);
    }

    /**
     * Returns the width of the visible part of the view in points.
     */
    final float getVisibleWidth() {
        return getWidth() * SCREENDPI / dpi;
    }

    /**
     * Returns the height of the visible part of the view in points.
     */
    final float getVisibleHeight() {
        return getHeight() * SCREENDPI / dpi;
    }

    /**
     * Returns the maximum distance that can be scrolled along the X-axis.
     * @return the maximum scroll distance for X
     */
    public final double getMaxScrollX() {
        return Math.max(pagewidth - getVisibleWidth(), 0);
    }

    /**
     * Returns the maximum distance that can be scrolled along the Y-axis.
     *
     * @return the maximum scroll distance for Y
     */
    public final double getMaxScrollY() {
        return Math.max(pageheight - getVisibleHeight(), 0);
    }

    /**
     * Set the {@link RenderingHints} to be used when rendering this PDF
     * @param hints the RenderingHints to use - may be null
     * @since 2.7.7
     */
    public void setRenderingHints(RenderingHints hints) {
        this.hints = hints;
    }

    /**
     * Sets the Viewports's {@link PDF} object, which can be <code>null</code>.
     * Calling this method resets the pagenumber, X and Y co-ordinates to 0, and the
     * DPI to 72.
     *
     * @param pdf the PDF pdf to display in this viewport
     * @see PDF
     */
    public final void setDocument(PDF pdf) {
        pageImage = fullImage = null;
        this.pdf = pdf;
        if (pdf == null) {
            parser = null;
            painters = null;
        } else {
            parser = new PDFParser(this.pdf);
            painters = new PagePainter[parser.getNumberOfPages()];
            for (int i=0;i<painters.length;i++) {
                painters[i] = parser.getPagePainter(i);
                painters[i].setRenderingHints(hints);
            }

            // Run the open action. If one doesn't exist, create one that
            // moves to the top-left of the first page.
            //
            PDFAction action = pdf.getAction(Event.OPEN);
            if (action==null) {
                PDFPage page = pdf.getPage(0);
                action = PDFAction.goTo(page, 0, page.getHeight(), 1);
            }
            runAction(action);
        }
    }

    /**
     * Select the page of the PDF to display in this Viewport. Calling this
     * method resets the X and Y co-ordinates to 0 but leaves the DPI unchanged.
     *
     * @param pagenumber the page number, between 0 and {@link PDF#getNumberOfPages}-1
     * @see #setDocument
     */
    public void setPageNumber(int pagenumber) {
        setPageNumber(pagenumber, 0, 0, 0);
    }

    /**
     * Display the specified page at the specified X and Y co-ordinates and DPI.
     *
     * @param pagenumber the page number, between 0 and {@link PDF#getNumberOfPages}-1
     * @param x the X co-ordinate of the top-left hand corner of the viewport, in points. NaN means "don't change"
     * @param y the Y co-ordinate of the top-left hand corner of the viewport, in points. NaN means "don't change"
     * @param dpi how many dots per inch to view the page at (the zoom level). A value of
     * 72 renders in 1 point per pixel, whereas a value of zero or NaN means keep the current zoom level.
     * @throws NullPointerException if no document has been set
     * @throws IllegalArgumentException if the DPI is <= 0
     */
    public void setPageNumber(int pagenumber, double x, double y, double dpi) {
        if (pdf==null) throw new NullPointerException("Document is null");
        if (dpi==dpi &&  dpi<0) throw new IllegalArgumentException("DPI must be > 0");
//        System.out.println("setPageNumber("+pagenumber+", "+x+", "+y+", "+dpi+")");

        if (pagenumber!=pagenum) {
            if (eventhandler!=null && pagenum!=-1) {
                PDFPage oldpage = pdf.getPage(pagenum);
                eventhandler.pageHidden(new PDFNavigationEvent(this, oldpage));
                removePageAnnotations(oldpage);
            }

            // Set "pagewidth" and "pageheight", the size of the page in points
            // Note if the CropBox is specified we use that, in which case pageleft
            // and pagebottom may be non-zero.
            //
            PDFPage page = pdf.getPage(pagenumber);
            float[] mbox = page.getBox("MediaBox");
            float[] cbox = page.getBox("CropBox");
            if (cbox == null) cbox = mbox;
            pageleft = cbox[0];
            pagebottom = cbox[1];
            pagewidth = cbox[2] - cbox[0];
            pageheight = cbox[3] - cbox[1];
            this.pagenum = pagenumber;
            page.flush();
            updateImage = true;

            if (eventhandler!=null) {
                storePageAnnotations(page);
                eventhandler.pageDisplayed(new PDFNavigationEvent(this, page));
            }
        }

        if (dpi==dpi && dpi!=0 && dpi!=this.dpi) {
            this.dpi = (float)dpi;
            updateImage = true;
        }

        // Set the view position and notify our observers that we've changed.
        // Allow for NaN for X or Y - means keep same position
        //
        if ((x==x && x!=x1) || (y==y && y!=y1) || updateImage) {
            setViewPosition(x, y);
            updateImage = true;
        }

        if (updateImage) {
            for (Iterator i = observers.iterator();i.hasNext();) {
                ((ViewportObserver)i.next()).viewportChanged(x1, y1);
            }
        }
    }

    /**
     * Sets the view co-ordinates that appear in the upper left hand corner
     * of the viewport.
     *
     * @param x the X co-ordinate of the top-left hand corner of the viewport, in points
     * @param y the Y co-ordinate of the top-left hand corner of the viewport, in points
     * @throws NullPointerException if no document has been set
     */
    public void setViewPosition(double x, double y) {
        if (pdf==null) throw new NullPointerException("Document is null");
//        System.out.print("setViewPosition("+x+", "+y+")");

        // Set 'x1" and "y1" to 0 if the full page is displayed, otherwise
        // set the section of the page to display
        //
        if (x==x) {
            x1 = getVisibleWidth() < pagewidth ? minmax(0, x, getMaxScrollX()) : 0;
            x2 = Math.min(x1+getVisibleWidth(), pagewidth);
        }
        if (y==y) {
            y1 = getVisibleHeight() < pageheight ? minmax(0, y, getMaxScrollY()) : 0;  // TODO?
            y2 = Math.min(y1+getVisibleHeight(), pageheight);
        }

        if (x2-x1!=0 && y2-y1!=0) {                     // If values aren't insane:
            // Adjust coordinates by crop offsets
            //
//            System.out.println(" page="+x1+" "+y1+" "+x2+" "+y2+" ph="+pageheight);

            if (task == null) {
                task = new PageWorker(pagenum, x1, y1, x2, y2, dpi, pageheight);
                task.start();
            } else {
                if ((task.x1!=x1 || task.y1!=y1 || task.x2!=x2 || task.y2!=y2 || task.dpi!=dpi || task.pagenum!=pagenum) && !task.isInterrupted()) {
                    task.interrupt();
                }
            }
        }
    }

    /**
     * Sets the zoom level of the viewport, in dots per inch.
     * @param dpi how many dots per inch to view the page at
     * @throws IllegalArgumentException if the DPI is <= 0
     */
    public void setDPI(double dpi) {
        setPageNumber(pagenum, x1, y1, dpi);
    }


    /**
     * Sets the view co-ordinates that appear in the upper left hand corner
     * of the viewport, and the dots per inch of the display area.
     *
     * @param x the X co-ordinate of the upper-left hand corner of the viewport, in points
     * @param y the Y co-ordinate of the upper-left hand corner of the viewport, in points
     * @param dpi how many dots per inch to view the page. A value of 72 gives in 1 point
     * per pixel. Must be > 0 or an {@link IllegalArgumentException} is thrown.
     *
     */
    public void setViewPosition(double x, double y, double dpi) {
        setPageNumber(pagenum, x, y, dpi);
    }

    /**
     * Returns the page number of the current {@link PDFPage} being displayed in the viewport.
     * @return the page number of the page currently displayed
     */
    public int getPageNumber() {
        return pagenum;
    }

    /**
     * Returns the view coordinates that appear in the upper left hand corner
     * of the viewport, or 0,0 if there's no view.
     */
    public Point2D getViewPosition() {
        return new Point2D.Double(x1, y1);
    }

    /**
     * Returns the dots per inch of the viewport's display area, as set by {link #setDPI}
     *
     * @return the dots per inch of the display area.
     */
    public double getDPI() {
        return dpi;
    }

    /**
     * Returns the width of the current {@link PDFPage} in points, rounded to the
     * nearest integer.  Returns 0 if no document has been set.
     *
     * @return the current width of the page in points
     */
    int getPageWidth() {
        return (int)pagewidth;
    }

    /**
     * Returns the height of the current {@link PDFPage} in points, rounded to the
     * nearest integer.  Returns 0 if no document has been set.
     *
     * @return the current height of the page in points
     */
    int getPageHeight() {
        return (int)pageheight;
    }

    /**
     * Convert the viewport's PDF to a TIFF image using the specified ColorModel
     * and dots per inch. The PDF is re-rendered in another thread, by calling
     * {@link PDFParser#writeAsTIFF}.
     *
     * @param out The OutputStream to write the TIFF to.
     * @param cm the ColorModel to use to render the images
     * @throws NullPointerException if no document as been set
     *
     * @see PDFParser#writeAsTIFF
     */
    public void writeAsTIFF(OutputStream out, int dpi, ColorModel cm) {
        if (pdf==null) {
            throw new NullPointerException("Document is null");
        }
        final OutputStream o = out;
        final int d = dpi;
        final ColorModel c = cm;
        new Thread () {
            public void run() {
                try {
                    parser.writeAsTIFF(o, d, c);
                    o.close();
                } catch (IOException io) {
                    ErrorInfo.displayThrowable(io, "TIFF Export Error", PDFViewport.this);
                }
                // notify observers
                Iterator iterator = observers.iterator();
                while (iterator.hasNext()) {
                    ((ViewportObserver) iterator.next()).writeAsTIFFComplete();
                }
            }
        }.start();
    }

    /**
     * Repaint the component.
     */
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        g.setColor(Color.GRAY);
        g.fillRect(0, 0, getWidth(), getHeight());
        if (pageImage != null) {
            int w = pageImage.getWidth(this);
            int h = pageImage.getHeight(this);
            int x = w < getWidth() ? (getWidth() - w) / 2 : 0;
            int y = h < getHeight() ? (getHeight() - h) / 2 : 0;

            // Drop shadow
            g.setColor(Color.BLACK);
            g.fillRect(x+3, y+3, w, h);

            // Image
            Shape clip = g.getClip();
            g.clipRect(x, y, w, h);
            g.drawImage(pageImage, x, y, this);

            // Transform the canvas, call the annotateComponent
            // method and then untransform it.
            AffineTransform f = ((Graphics2D)g).getTransform();
            AffineTransform newf = new AffineTransform(dpi/SCREENDPI, 0, 0, -dpi/SCREENDPI,
                                                        -x1*dpi/SCREENDPI + x,
                                                        (pageheight-y1)*dpi/SCREENDPI + y);
            ((Graphics2D)g).transform(newf);
            annotateComponent((Graphics2D)g);
            ((Graphics2D)g).setTransform(f);

            g.setClip(clip);
            g.setColor(Color.BLACK);
            g.drawRect(x, y, w, h);
        }
    }

    /**
     * This method is called whenever the PDF page is redrawn. By default it
     * does nothing, but it can be overridden by subclasses to annotate the
     * canvas. The co-ordinate system in use when this method is called is
     * the PDF co-ordinate system, so (0,0) is the bottom left of the page
     * and the DPI, scroll position etc. of this PDFViewport can be ignored.
     * </p><p>
     * More advanced applications can use the tranformation applied to the
     * Graphics2D object to reposition any components this Viewport may have.
     * The following code would get you started:
     * <pre class="example">
     *  AffineTransform tran = g.getTransform();
     *  final float scalex = (float)tran.getScaleX();
     *  final float scaley = (float)tran.getScaleY();
     *  final float dx = (float)tran.getTranslateX();
     *  final float dy = (float)tran.getTranslateY();
     *
     *  Component[] comps = getComponents();
     *  for (int i=0;i&lt;comps.length;i++) {
     *      Rectangle r = getComponentPosition(comps[i]);
     *      float x = r.x*scalex + dx;
     *      float y = r.y*scaley + dy;
     *      float w = r.width * scalex;
     *      float h = r.height * Math.abs(scaley);
     *      comps[i].setBounds((int)x, (int)y, (int)w, (int)h);
     *  }
     * </pre>
     *
     * @param g the Graphics2D object containing the just-rendered page.
     * @since 2.7.1
     */
    protected void annotateComponent(Graphics2D g) {
    }

    /**
     * Adds a {@link ViewportObserver} to be notified of viewport events.
     * @param observer the {@link ViewportObserver} to be added
     */
    public void addObserver(PDFViewport.ViewportObserver observer) {
        observers.add(observer);
    }

    /**
     * Removes a {@link ViewportObserver}.
     * @param observer the {@link ViewportObserver} to be removed
     */
    public void removeObserver(PDFViewport.ViewportObserver observer) {
        observers.remove(observer);
    }

    /**
     * Displays a print dialog for printing any pages from the document in the viewport. Print
     * a single page or a range of pages.
     *
     * @throws NullPointerException if no document has been set
     * @see #setDocument(PDF)
     */
    public void print() {
        if (pdf==null) throw new NullPointerException("Document is null");

        DocFlavor flavor = DocFlavor.SERVICE_FORMATTED.PAGEABLE;
        PrintService[] services = PrintServiceLookup.lookupPrintServices(flavor, null);

        if (services.length > 0) {
            try {
                PrinterJob job = PrinterJob.getPrinterJob();
                job.setPageable(parser);
                if (pdf.getInfo("Title")!=null) {
                    job.setJobName(pdf.getInfo("Title"));
                } else {
                    job.setJobName("BFO PDF");
                }

                if(job.printDialog()) {
                    job.print();
                }

            } catch (PrinterException e) {
                e.printStackTrace();
            }
        } else {
            JOptionPane.showMessageDialog(this, "No printers are installed","Print Warning",JOptionPane.WARNING_MESSAGE);
        }
    }

    class PageWorker extends SwingWorker {
        private boolean interrupted;
        final int pagenum;
        final float dpi, x1, y1, x2, y2, pageheight;

        PageWorker(int pagenum, float x1, float y1, float x2, float y2, float dpi, float pageheight) {
            this.pageheight = pageheight;
            this.pagenum = pagenum;
            this.dpi = dpi;
            this.x1 = x1;
            this.x2 = x2;
            this.y1 = y1;
            this.y2 = y2;
        }

        /**
         * Here we recreate pageImage.
         *
         * 1. If the zoom level and image size are such that we can render the
         *    entire page in a "reasonable" size, we render the entire page to
         *    "fullImage" and then createa a subImage of that as "pageImage".
         *    This is faster than it sounds as the raster is not copied.
         *
         * 2. If not, we render only the portion we can to "pageImage"
         */
        public Object construct() {
            try {
                if (!isFullImageOK()) {
                    pageImage = painters[pagenum].getSubImage(pageleft+x1, pagebottom+pageheight-y2, pageleft+x2, pagebottom+pageheight-y1, dpi, colormodel);
                } else {
                    if (updateImage) {
                        // page or dpi changed so render full page
                        fullImage = painters[pagenum].getSubImage(pageleft, pagebottom, pageleft+pagewidth, pagebottom+pageheight, dpi, colormodel);
                    }
                    if (fullImage != null) {
                        float x = x1 * dpi / SCREENDPI;
                        float y = y1 * dpi / SCREENDPI;
                        float w = fullImage.getWidth() - x;
                        float h = fullImage.getHeight() - y;
//                        System.out.println("fw="+fullImage.getWidth()+" fh="+fullImage.getHeight()+" x="+x+" y="+y+" w="+w+" h="+h);
                        pageImage = fullImage.getSubimage((int)x, (int)y, (int)w, (int)h);
                    }
               }
            } catch (Exception e) {
                fullImage = null;
                pageImage = null;
                ErrorInfo.displayThrowable(e, "Page Painter Error", PDFViewport.this);
            }
            return null;
        }

        public void finished() {
            if (task.isInterrupted()) {
                task = null;
                setViewPosition(PDFViewport.this.x1, PDFViewport.this.y1);
            } else {
                task = null;
                updateImage = false;
                repaint();
            }
        }

        public void interrupt() {
            interrupted = true;
            painters[pagenum].interrupt();
        }

        boolean isInterrupted() {
            return interrupted;
        }
    }

    /**
     * This interface needs to be implemented by any class that wants to
     * be notified of {@link PDFViewport} events. For a <code>ViewportObserver</code>
     * to recieve events they need to add themselves as an observer.
     *
     * @see PDFViewport#addObserver(PDFViewport.ViewportObserver)
     */
    public interface ViewportObserver {
        /**
         * This event fires whenever the {@link PDFViewport} changes the page
         * or the DPI.
         *
         * @param x the X co-ordinate of the upper-left hand corner of the viewport, in points
         * @param y the Y co-ordinate of the upper-left hand corner of the viewport, in points
         * @see PDFViewport
         */
        public void viewportChanged(double x, double y);

        /**
         * This event fires whenever the {@link PDFViewport#writeAsTIFF writeAsTIFF}
         * event is completed.
         * @see PDFViewport
         */
        public void writeAsTIFFComplete();
    }

    //----------------------------------------------------------------------------
    // Methods relating to interaction follow

    /**
     * Set the event handler for any mouse clicks. The default handler
     * allows some basic interaction with annotations - hyperlink processing
     * and the like.
     * @param handler the {@link EventHandler} that should be used to handle
     * mouse or keyboard events in this viewport.
     * @since 2.7.8
     */
    public void setEventHandler(EventHandler handler) {
        if (handler!=null && handler.getViewport()!=null && handler.getViewport()!=this) {
            throw new IllegalStateException("EventHandler already registered with a viewport");
        }

        boolean oldmouse = eventhandler==null ? false : (eventhandler.getType()&handler.TYPE_MOUSE)!=0;
        boolean oldmotion = eventhandler==null ? false : (eventhandler.getType()&handler.TYPE_MOUSEMOTION)!=0;
        boolean oldkey = eventhandler==null ? false : (eventhandler.getType()&handler.TYPE_KEYBOARD)!=0;

        boolean newmouse = handler==null ? false : (handler.getType()&handler.TYPE_MOUSE)!=0;
        boolean newmotion = handler==null ? false : (handler.getType()&handler.TYPE_MOUSEMOTION)!=0;
        boolean newkey = handler==null ? false : (handler.getType()&handler.TYPE_KEYBOARD)!=0;

        if (newmouse && !oldmouse) {
            addMouseListener(this);
        } else if (!newmouse && oldmouse) {
            removeMouseListener(this);
        }
        if (newmotion && !oldmotion) {
            addMouseMotionListener(this);
        } else if (!newmotion && oldmotion) {
            removeMouseMotionListener(this);
        }
        if (newkey && !oldkey) {
            addKeyListener(this);
        } else if (!newkey && oldkey) {
            removeKeyListener(this);
        }

        if (eventhandler!=null) eventhandler.setViewport(null);
        eventhandler = handler;
        if (eventhandler!=null) {
            eventhandler.setViewport(this);
            if (pagenum!=-1) {
                eventhandler.pageDisplayed(new PDFNavigationEvent(this, pdf.getPage(pagenum)));
            }
        }
    }

    /**
     * Run the specified action on the PDF. Currently only the "GoTo" actions are
     * implemented, plus a few basic Named actions.
     * @param action the Action to run.
     * @return true if the action was recognised and run successfully, false otherwise.
     * @since 2.7.7
     */
    public boolean runAction(PDFAction action) {
        boolean success = false;
        while (action!=null) {
            success = true;
            PDFPage page = null;
            float x = Float.NaN, y = Float.NaN, zoom = 0;

            String type = action.getType();
            if (type.equals("GoToFit")) {
                page = action.getPage();
                zoom = Math.min((float)getWidth()/page.getWidth(), (float)getHeight()/page.getHeight());
            } else if (type.equals("GoTo")) {
                page = action.getPage();
                float[] coords = action.getGoToCoordinates();
                x = coords[0];
                y = coords[1];
                zoom = coords[2];
            } else if (type.equals("GoToFitWidth")) {
                page = action.getPage();
                float[] coords = action.getGoToCoordinates();
                zoom = (float)getWidth()/page.getWidth();
                y = coords[0];
            } else if (type.equals("GoToFitHeight")) {
                page = action.getPage();
                float[] coords = action.getGoToCoordinates();
                zoom = (float)getHeight()/page.getHeight();
                x = coords[0];
            } else if (type.equals("GoToFitRectangle")) {
                page = action.getPage();
                float[] coords = action.getGoToCoordinates();
                x = Math.min(coords[0], coords[2]);
                y = Math.min(coords[1], coords[3]);
                zoom = Math.min(getWidth() / Math.abs(coords[2]-coords[0]), getHeight() / Math.abs(coords[3]-coords[1]));

            } else if (type.equals("Named:NextPage")) {
                if (pagenum < pdf.getNumberOfPages()-2) {
                    page = pdf.getPage(pagenum+1);
                }
            } else if (type.equals("Named:PrevPage")) {
                if (pagenum > 0) {
                    page = pdf.getPage(pagenum-1);
                }
            } else if (type.equals("Named:FirstPage")) {
                page = pdf.getPage(0);
            } else if (type.equals("Named:LastPage")) {
                page = pdf.getPage(pdf.getNumberOfPages()-1);
            } else if (type.equals("Named:Print")) {
                print();
            } else {
                success = false;
//                System.out.println("Action="+action);
            }

            if (page!=null) {
                if (!pdf.getPages().contains(page)) {
                    throw new IllegalArgumentException("Action refers to a page from another PDF");
                }
                setPageNumber(page.getPageNumber()-1, x, page.getHeight()-y, zoom*SCREENDPI);
            }
            action = action.getNext();
        }
        return success;
    }

    /**
     * Return the annotation this mouse event was on. We iterate through each
     * shape stored in the annotationmap (as created by storePageAnnotation), and
     * check if it contains the point. Returns null if no annotations match.
     *
     * Not very efficient, should probably store these as a B-Tree.
     */
    PDFAnnotation getAnnotation(PDFMouseEvent event) {
        PDFPage page = event.getPage();
        Map hotspots = (Map)annotationmap.get(page);
        if (hotspots!=null) {
            Point2D point = event.getPagePoint();
            for (Iterator i = hotspots.entrySet().iterator();i.hasNext();) {
                Map.Entry e = (Map.Entry)i.next();
                Shape s = (Shape)e.getKey();
                if (s.contains(point)) {
                    return (PDFAnnotation)e.getValue();
                }
            }
        }
        return null;
    }

    /**
     * Store the shapes for each annotation on the specified page in the "annotationsmap"
     * map. The key is a Shape, the value a PDFAnnotation
     */
    private void storePageAnnotations(PDFPage page) {
        Map map = (Map)annotationmap.get(page);
        if (map==null) {
            map = new HashMap();
            annotationmap.put(page, map);
        }
        for (Iterator i = page.getAnnotations().iterator();i.hasNext();) {
            PDFAnnotation annot = (PDFAnnotation)i.next();
            Shape shape = null;
            if (!(annot instanceof AnnotationMarkup) && !(annot.getType().equals("Popup"))) {
                if (annot instanceof AnnotationLink) {
                    float[] c = ((AnnotationLink)annot).getCorners();
                    if (c!=null) {
                        GeneralPath p = new GeneralPath();
                        p.moveTo(c[0], c[1]);
                        p.lineTo(c[2], c[3]);
                        p.lineTo(c[4], c[5]);
                        p.lineTo(c[6], c[7]);
                        p.closePath();
                        shape = p;
                    }
                }
                if (annot instanceof AnnotationNote && !((AnnotationNote)annot).isOpen()) {
                    float[] c = annot.getRectangle();
                    shape = new Rectangle2D.Float(c[0], c[1], 20, 20);
                }
                if (shape==null) {
                    float[] c = annot.getRectangle();
                    shape = new Rectangle2D.Float(c[0], c[1], c[2]-c[0], c[3]-c[1]);
                }
            }
            if (shape!=null) {
                map.put(shape, annot);
            }
        }
    }

    private void removePageAnnotations(PDFPage page) {
        annotationmap.remove(page);
    }

    //----------------------------------------------------------------------
    // Glue for event handling

    /**
     * Given a regular AWT MouseEvent, wrap it in a PDFMouseEvent and return. This
     * only happens if the event occurred on a page - if not, return null.
     */
    private PDFMouseEvent getMouseEvent(MouseEvent event) {
        int framex = pageImage==null ? 0 : pageImage.getWidth(this) < getWidth() ? (getWidth() - pageImage.getWidth(this)) / 2 : 0;
        float pagex = (float)(((event.getX() - framex) / getDPI() * 72f) + x1);

        int framey = pageImage==null ? 0 : pageImage.getHeight(this) < getHeight() ? (getHeight() - pageImage.getHeight(this)) / 2 : 0;
        float pagey  = (float)(pageheight - y1 - ((event.getY()-framey) / getDPI() * 72f));

        if (pagex >= 0 && pagex <= pagewidth && pagey>=0 && pagey <= pageheight) {
            PDFPage page = pdf.getPage(pagenum);
            return new PDFMouseEvent(event, this, page, pagex, pagey);
        } else {
            return null;
        }
    }

    /**
     * Given a regular AWT KeyEvent, wrap it in a KeyEvent and return.
     */

    private PDFKeyEvent getKeyEvent(KeyEvent event) {
        return new PDFKeyEvent(event, this);
    }
    public void mouseEntered(MouseEvent event) {
    }
    public void mouseExited(MouseEvent event) {
    }
    public void mousePressed(MouseEvent event) {
        PDFMouseEvent pdfe = getMouseEvent(event);
        if (pdfe!=null) eventhandler.mousePressed(pdfe);
    }
    public void mouseReleased(MouseEvent event) {
        PDFMouseEvent pdfe = getMouseEvent(event);
        if (pdfe!=null) eventhandler.mouseReleased(pdfe);
    }
    public void mouseMoved(MouseEvent event) {
        PDFMouseEvent pdfe = getMouseEvent(event);
        if (pdfe!=null) eventhandler.mouseMoved(pdfe);
    }
    public void mouseDragged(MouseEvent event) {
        PDFMouseEvent pdfe = getMouseEvent(event);
        if (pdfe!=null) eventhandler.mouseDragged(pdfe);
    }
    public void mouseClicked(MouseEvent event) {
        PDFMouseEvent pdfe = getMouseEvent(event);
        if (pdfe!=null) eventhandler.mouseClicked(pdfe);
    }
    public void keyPressed(KeyEvent event) {
        PDFKeyEvent pdfe = getKeyEvent(event);
        if (pdfe!=null) eventhandler.keyPressed(pdfe);
    }
    public void keyReleased(KeyEvent event) {
        PDFKeyEvent pdfe = getKeyEvent(event);
        if (pdfe!=null) eventhandler.keyReleased(pdfe);
    }
    public void keyTyped(KeyEvent event) {
        PDFKeyEvent pdfe = getKeyEvent(event);
        if (pdfe!=null) eventhandler.keyTyped(pdfe);
    }
}
